;; Generated package description from doc-view-follow.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "doc-view-follow" "0.3.2" "Synchronize windows showing the same document" '((emacs "29.1")) :commit "0393bfc9cdbec698201c9c69c247b201558f8511" :authors '(("Paul D. Nelson" . "ultrono@gmail.com")) :maintainer '("Paul D. Nelson" . "ultrono@gmail.com") :keywords '("convenience") :url "https://github.com/ultronozm/doc-view-follow.el")
